/*  Module Name: @(#)api_usmarc_formatter.c	7.2
    Version: 3/5/97
    Author: John Hough
 
    Crossnet Systems Ltd.
    Unit 3, The Paddock,
    Hambridge Road,
    Newbury,
    Berkshire, RG14 5TQ
 
    Date of Issue:
 
    QA Acceptance:
    Date of Acceptance:
 
    Description: ASCII USMARC formatter.
		 The 'unofficial' standard for Z39.50 V2 records 
		 is ASCII USMARC over TCP/IP - nothing else. In the
		 Z39.50 world, ASCII USMARC rules. Inspired by code in the 
	  	 Willow Z39.50 application.
		 This code may be freely used/modified subject to the terms 
		 and conditions of the GNU general public licence.


*/


#include "api_usmarc_formatter.h"


/*********************************************************************
/* usmarc_format - Routine for formatting a DbvOctetString 
		   containing USMARC data (ascii, USMARC is the unofficial
		   Z39.50 V2 standard)

Input parameters:
	psDbvOctetstring -	Pointer to DbvOctetstring structure containing
			  	USMARC data

	eMarcType -		Type of formatting required (0 [Tomas_marc] or
			   	1 [Normal_marc])

	ePrintHeader -		Header printing on/off (0 [DBV_FALSE] or 
				1 [DBV_TRUE])

	file -		   	Where to send formatting results and diagnostics 


Output parameters: 
	None


Examples of usage:

DbvOctetString mystruct;

usmarc_format(&mystruct, Tomas_marc, DBV_TRUE, stdout);
usmarc_format(&mystruct, Normal_marc, DBV_TRUE, stdout);
***********************************************************************/


int usmarc_format(DbvOctetString *psDbvOctetString, MarcType eMarcType, DbvBool ePrintHeader, FILE *file)
{
	char *record = NULL;
	int length = 0;


	/* Access the DbvOctetString elements */
	length = (int) psDbvOctetString->ulLength;
	record = psDbvOctetString->pcValue;

	/* If holdings record */
	if (holdings_rec(record))
            hold_rec(record, length, file);   

	/* else have a bibliographical record */
	else {
		switch (eMarcType) {

		/* If to be formatted as TOMAS USMARC */
		case Tomas_marc:
            		bib_rec(record, length, eMarcType, ePrintHeader, file);
			break;

		/* If to be formatted as normal USMARC (TOMAS D/DN) */
		case Normal_marc:
            		bib_rec(record, length, eMarcType, ePrintHeader, file);
			break;
		}
	}

	return 1;
}



/* Returns 1 if holdings record */
int holdings_rec(char *record)
{
	if (record[6] == 'y')          
		return 1;

	return 0;
}



int hold_rec(char *record, int length, FILE *f)  
{
	if (check_hold_length(length, f) == -1)
		return -1;

	fprintf(f, "Holdings record formatting not supported\n");
	return -1;
}



int check_hold_length(int length, FILE *f)
{
	/* Check length */
	if (length < MINHLEN || length > MAXHLEN ) {
		fprintf(f,"Holdings record length of %d is invalid\n", length);
		return -1;
	}

	return 1;
} 



int bib_rec(char *record, int length, MarcType eMarcType, DbvBool ePrintHeader, FILE *f)
{
	int addr, p;

       	int code, offset, compon_length, data_start, data_end;

	int record_count = 0;
	
	if (check_bib_length(length, f) == -1)
		return -1;

	addr = int_from_string(record+12, 5, f);  

	if (check_addr(addr, length, f) == -1)
		return -1;


	/* if Header stuff (chars 0-23 inclusive) to be printed ? */

	if (ePrintHeader)
		print_bib_leader(record, f);    

	/* Data starts at char 24 */
	p = 24;                             
	while ( p < addr && record[p] != REC_SEP ) {

		if (get_code_length_offset(record, &code, &compon_length, &offset, &p, f) == -1)
			return -1;

			data_start = addr+offset;
			data_end = addr+offset+compon_length-1; 

        		if (data_end > length) 
            			fprintf(f, "Error in data length"); 

			switch (eMarcType) {

			/* If to be formatted as TOMAS USMARC */
			case Tomas_marc:
    				record_count++;
    				fprintf(f, "%3d %03d ", record_count, code );
    				print_data(record+data_start, compon_length, f);     
    				fprintf(f, "\n");                      

				break;

			/* If to be formatted as normal USMARC (TOMAS D/DN) */
			case Normal_marc:
				ndisplay(code, record, data_start, compon_length, f);
				break;
			}

	}
	return 1;
} 



int check_bib_length(int length, FILE *f)
{
	if (length < MINBIBLEN || length > MAXBIBLEN ) {
		fprintf(f,"Bibliographical record length of %d is invalid\n", length);
		return -1;
	}

	return 1;
}



/* Get a fixed compon_length integer */
int int_from_string(char *string, int len, FILE *f)
{
    int c, i = 0;

    while( len-- ) {
        c = *string++;
        if (c < '0' || c > '9')
            fprintf(f,"Invalid digit %c", c);
        i = (i * 10) + c - '0';
    } 

    return i;
} 



int check_addr(int addr, int length, FILE *f)
{
	if (addr < 24 || addr > length) {
        	fprintf(f,"Invalid address %d (length %d)", addr, length);
		return -1;
	}
	return 1;
}



/* Print bib record leader */
int print_bib_leader(char *record, FILE *f)                  
{
	/* Record status */
	print_rec_status(record, f);

	fprintf(f, "Record type: %c\n", record[6] );
 	fprintf(f, "Bibliographical level: %c\n", record[7] );

  	if (check_8_9_10_11(record, f) == -1)
		return -1;
		
 	/* Characters 12-16 inclusive = address */

	fprintf(f, "Encoding level: %c\n", record[17] );
	fprintf(f, "Description catalogging form: %c\n", record[18] );

	/* Character 19 blank */
	/* Character 20 holds compon_length */
	/* Character 21 holds offset */

	/* Print transaction type */
	print_trans_type(record, f);

	/* Character 23 must be '0' */

	return 1;
} 


int print_rec_status(char *record, FILE *f)
{
	fprintf(f, "Record status: ");
	switch( record[5] ) {
	case 'n':
        	fprintf(f, "new\n");
        	break;
    	case 'c':
        	fprintf(f, "corrected\n");
        	break;
    	case 'p':
        	fprintf(f, "prepublication\n");
        	break;
    	case 'a':
        	fprintf(f, "coding level increase\n");
        	break;
    	default:
        	fprintf(f, "unknown (%c)\n", record[5] );
        	break;
    } 
}



int check_8_9_10_11(char *record, FILE *f)
{
	/* No error by default */
	int error_pos = 0;

	if (record[8] != ' ') 
		error_pos = 8;

	if (record[9] != ' ') 
		error_pos = 9;

	if (record[10] != '2') 
		error_pos = 10;

	if (record[11] != '2') 
		error_pos = 11;

	if (error_pos != 0) {
        	fprintf(f,"Invalid contents character in bibliographical record %d\n", error_pos);
		return -1;
	}
		
	return 1;
}



int print_trans_type(char *record, FILE *f)
{
	/* Transaction type */

	fprintf(f, "Transaction type: ");
	switch( (unsigned char) record[22] ) {
	case 0x01:
		fprintf(f, "produce");
		break;
	case 0x02:
		fprintf(f, "update");
		break;
	case 0x03:
		fprintf(f, "cancel update");
		break;
	case 0x11:
		fprintf(f, "replace");
		break;
	case 0x50:
		fprintf(f, "all produce");
		break;
	case 0x90:
		fprintf(f, "offline retrieve");
		break;
	case 0x92:
		fprintf(f, "offline update");
		break;
	case 0x93:
		fprintf(f, "offline cancel update");
		break;
	case 0x94:
		fprintf(f, "microcon update");
		break;
	default:
		fprintf(f, "unknown\n");
		break;
	}
} 



int get_code_length_offset(char *record, int *code, int *compon_length, int *offset, int *p, FILE *f)
{
	int lengthsiz, offsiz;

       	*code = int_from_string( record + (*p), CODELEN, f );
       	*p += CODELEN;

	/* Number of columns in compon_length */
       	lengthsiz = record[20] - '0';      
        if (lengthsiz != 4 )
		fprintf(f,"%03d: invalid compon_length field size (column 20): %d", code, lengthsiz );

	/* Get compon_length */
       	*compon_length = int_from_string( record + (*p), lengthsiz, f); 
       	*p += lengthsiz;                    

	/* Number of columns for offset */
       	offsiz = record[21] - '0';      
       	if (offsiz != 5 )
		fprintf(f,"%03d: unusual size of offset field (column 21): %d", code, offsiz );

	/* Get offset */
       	*offset = int_from_string( record + (*p), offsiz, f); 
       	*p += offsiz; 

	return 1;
} 



/* Print data */
int print_data(char *dp, int length, FILE *f)
{
	char c;

	while (length-- > 0) {

		/* Next char */
        	c = *dp++; 

		/* If ('\036') field seperator */
        	if (c == REC_SEP )                  
            		break;

		/* If ('\037') start of subsection */
		if (c == REC_SUBSEC ) { 

			/* Account for subsection letter */
           		length--;                      

			/* If no more data */
            		if( length < 0 )               
                		fprintf(f,"Bad subfield");  

            		/* print next char (sub section letter) */
            		fprintf(f, " %c%c ", SCHAR, *dp++); 
            		continue; 
		} 

	if (isascii(c))
            fprintf(f, "%c", c );               
    } 
} 




/**********************************************************/


/* normal "marc" format (TOMAS D/DN) */

int ndmarc(char* record, int length, FILE *f)
{
	/* If holdings_rec record */
	if (holdings_rec(record))
            hold_rec(record, length, f);   
        else
            nbib_rec(record, length, f);
} 



int nbib_rec(char *record, int length, FILE *f)
{
	int addr, p;

       	int code, offset, compon_length, data_start, data_end;

	
	if (check_bib_length(length, f) == -1)
		return -1;

	addr = int_from_string(record+12, 5, f);  

	if (check_addr(addr, length, f) == -1)
		return -1;

	/* Print leader 24 characters (0 - 23) */
	print_bib_leader(record, f);    

	/* Data starts at char 24 */
	p = 24;                             
	while ( p < addr && record[p] != REC_SEP ) {

		if (get_code_length_offset(record, &code, &compon_length, &offset, &p, f) == -1)
			return -1;

			data_start = addr+offset;
			data_end = addr+offset+compon_length-1; 

        		if ( data_end > length ) 
            			fprintf(f, "data out of bounds"); 

			ndisplay(code, record, data_start, compon_length, f);
	}
	return 1;
} 






/****************************************************************/



ndisplay( int code, char *record, int offset, int compon_length, FILE *f)
{

	
    char bib_lvl;

	str auth, title, publ, descr, notes, subj, other, loc, call;
	int authtype, titletype, publtype, descrtype, notestype,
	subjtype, othertype, loctype,  calltype;

    authtype = titletype = publtype = descrtype = notestype =
        subjtype = othertype = loctype = calltype = -1;

	/* Skip 2 indicators */
    record += offset+2;   


    switch( code ) {
    case 49:
        if( loctype != -1 )
            fprintf(f, "second location");
        copy( loc, record, compon_length );
        loctype = code;
        break;

    case 50:
    case 90:
    case 99:
        if( calltype != -1 )
            fprintf(f, "second call %03d (was %03d)\n", code, calltype );
        copy( call, record, compon_length );
        calltype = code;
        break;


    case 100:
    case 110:
    case 130:
        if( authtype != -1 )
            fprintf(f, "second author %03d (was %03d)\n", code, authtype );
        copy( auth, record, compon_length );
        authtype = code;
        break;

    case 240:
    case 245:
        if( titletype != -1 )
            fprintf(f, "second title %03d (was %03d)\n", code, titletype );
        copy( title, record, compon_length );
        titletype = code;
        break;

    case 260:
        if( publtype != -1 )
            fprintf(f, "second publication");
        copy( publ, record, compon_length );
        publtype = code;
        break;

    case 300:
        if( descrtype != -1 )
            fprintf(f, "second description");
        copy( descr, record, compon_length );
        descrtype = code;
        break;

    } 

    show_field(auth, authtype,  "AUTHOR", f);
    show_field(title,titletype, "TITLE", f);
    show_field(publ, publtype,  "PUBLICATION", f );
    show_field(descr,descrtype, "DESCRIPTION", f );
    show_field(notes,notestype, "NOTES", f );          
    show_field(subj, subjtype,  "SUBJECTS", f );       
    show_field(other,othertype, "OTHER", f );          
    show_field(loc,  loctype,   "LOCATION", f );       
    show_field(call, calltype,  "CALL NUMBER", f );    
} 




int copy(char *dp, char *sp, int length )
{
    while( length-- > 0 ) {
        int c;

	/* Get next character */
        c = *sp++;                      

	/* REC_SEP ? */
        if( c == REC_SEP )                   
            break;                      

	/* subfield delimiter ? */
        else if( c == REC_SUBSEC ) {            

		/* account for subsection letter */
            length--;                      

		/* no more data ? */
            if( length < 0 )               
                fprintf(stderr,"bad subfield");   

	    /* Ignore section letter */
            sp++;                       

	/* Insert space */
            *dp++ = ' ';                

        } 

	else if (isascii(c))
            *dp++ = c;
    } 
    *dp = EOS;
} 


show_field(char *data, int type, char *name, FILE *f)
{
    if (type != -1 ) 
	fprintf(f, "%-16s %s\n", name, data ); 
	
}



