/*  Module Name: @(#)shared_z3950_close.c	7.1
    Version: 11/15/96
    Author: John Hough
    
    Crossnet Systems Ltd.
    Unit 3, The Paddock,
    Hambridge Road,
    Newbury,
    Berkshire, RG14 5TQ

    Date of Issue:  

    QA Acceptance:  
    Date of Acceptance:  

    Description: Origin/Target shared routines to create a Close PDU to be
		 sent to the other end of the connection and process
		 a received Close PDU sent by the other end of the connection
		 Note: as that there is only one Close PDU in z39.50,
		 DbvCloseRequest() and DbvCloseResponse() both call 
		 DBVClose() to setup a CloseReq PDU.


*/

#include "shared_z3950_close.h"

/*****************************************************************
******************************************************************
******************************************************************
CREATE Close PDU TO SEND TO Other End Of Connection AND SEND IT
*/


int DbvCloseRequest(DbvAssocId lAssocId, DbvCloseReq *psCloseReq, DbvError **ppsSIError)
{
	int return_value = SI_OK;

	/* enumerated integer type: O = origin, T = target */
	DBVOriginOrTarget sender = O; 
	DBVAssociateDetails *psAssociateDetails;
	DbvProtocolStack eProtocolStack;
	DBVCloseType closetype = CL_Req;


	/* If fields invalid */
	if (DBVCheckDbvCloseReqFields(lAssocId, psCloseReq, ppsSIError) == SI_NOTOK)
		return SI_NOTOK;

	/* If stack invalid */
	if (SIWhichStack(lAssocId, &eProtocolStack, ppsSIError) == SI_NOTOK)
		return SI_NOTOK;

	psAssociateDetails = SL_Current(DBVAssociateList);
	sender = psAssociateDetails->eOriginOrTarget;
	

	DBVEnterFuncTrace("DBVCloseRequest",lAssocId);

	/* A closeRequest can be sent from origin to target or visa versa */

	/* If origin */
	if (sender == O)
		if (DBVChangeState(lAssocId, DBV_OE_CLOSEREQ, ppsSIError) == SI_OK)
			return_value = DBVClose(lAssocId, psCloseReq, closetype, ppsSIError);
		else
			return_value = SI_NOTOK;

	/* If target */
	if (sender == T)
		if (DBVChangeState(lAssocId, DBV_TE_CLOSEREQ, ppsSIError) == SI_OK) 
			return_value = DBVClose(lAssocId, psCloseReq, closetype, ppsSIError);
		else
			return_value = SI_NOTOK;

	DBVLeaveFuncTrace("DBVCloseRequest",lAssocId, return_value);

	return return_value;
}



/* There is only one Close PDU in Z39.50 */
int DbvCloseResponse(DbvAssocId lAssocId, DbvCloseReq *psCloseRsp, DbvError **ppsSIError)
{
	int return_value = SI_OK;

	/* enumerated integer type: O = origin, T = target */
	DBVOriginOrTarget sender = O; 
	DBVAssociateDetails *psAssociateDetails;
	DbvProtocolStack eProtocolStack;
	DBVCloseType closetype = CL_Rsp;


	if (DBVCheckDbvCloseRspFields(lAssocId, psCloseRsp, ppsSIError) == SI_NOTOK)
		return SI_NOTOK;

	if (SIWhichStack(lAssocId, &eProtocolStack, ppsSIError) == SI_NOTOK)
		return SI_NOTOK;

	psAssociateDetails = SL_Current(DBVAssociateList);
	sender = psAssociateDetails->eOriginOrTarget;
	

	DBVEnterFuncTrace("DBVCloseResponse",lAssocId);

	/* If origin */
	if (sender == O)
		if (DBVChangeState(lAssocId, DBV_OE_CLOSERESP, ppsSIError) == SI_OK)
			return_value = DBVClose(lAssocId, psCloseRsp, closetype, ppsSIError);
		else
			return_value = SI_NOTOK;

	/* If target */
	if (sender == T)
		if (DBVChangeState(lAssocId, DBV_TE_CLOSERESP, ppsSIError) == SI_OK)
			return_value = DBVClose(lAssocId, psCloseRsp, closetype,  ppsSIError);
		else
			return_value = SI_NOTOK;


	DBVLeaveFuncTrace("DBVCloseResponse",lAssocId, return_value);

	return return_value;
}



int DBVClose(DbvAssocId lAssocId, DbvCloseReq *psCloseReq, DBVCloseType closetype, DbvError **ppsSIError)
{

	PDU pdu_to_encode;
	PDU *PDU_ptr;
	DbvProtocolStack eProtocolStack;
	int return_value = SI_NOTOK;
	int val;
	jmp_buf env;

	return_value = SI_NOTOK;

	/* Field checking and state changing done by calling functions */ 
	if (SIWhichStack(lAssocId, &eProtocolStack, ppsSIError) == SI_OK)

		/* PDU can be encoded and sent */
		return_value = SI_OK;

	if(return_value == SI_OK)
	{
		DBVSendingTrace(lAssocId);

		/* If 'Close Request' */
		if (closetype == CL_Req)
                	DBVCloseReqTrace(lAssocId, psCloseReq);
		else
                	DBVCloseRspTrace(lAssocId, psCloseReq);

		PDU_ptr = &pdu_to_encode;
		if ((val = setjmp(env)) == 0) {
			PDU_ptr->a.close = (struct Close *) Asn1Alloc (sizeof(struct Close));
			CheckAsn1Alloc(PDU_ptr->a.close, env);
			PDU_ptr->choiceId = PDU_CLOSE;
			DBVsetup_snacc_closerequest(PDU_ptr, psCloseReq, env);
		} else {
			DBVfill_error_structure(lAssocId, SI_Interface_Error, 0,DBV_PDU_OUTOFMEM, DBV_PDU_OUTOFMEM_S, 0, Association_Open, ppsSIError);
                        return SI_NOTOK;
                }

		return_value = DBVencode_and_send(PDU_ptr, lAssocId, ppsSIError);
	}

	return return_value;
}


int DBVCheckDbvCloseReqFields(DbvAssocId lAssocId, DbvCloseReq *psCloseReq, DbvError **ppsSIError)
{
	if (DBVCheckDbvCloseFields(lAssocId, psCloseReq, "psCloseReq->", ppsSIError) == SI_NOTOK)
		return SI_NOTOK;

	return SI_OK;

}



int DBVCheckDbvCloseRspFields(DbvAssocId lAssocId, DbvCloseReq *psCloseRsp, DbvError **ppsSIError)
{
	if (DBVCheckDbvCloseFields(lAssocId, psCloseRsp, "psCloseRsp->", ppsSIError) == SI_NOTOK)
		return SI_NOTOK;

	return SI_OK;
}



int DBVCheckDbvCloseFields(DbvAssocId lAssocId, DbvCloseReq *psCloseReq, char *error_buffer, DbvError **ppsSIError)
{
	char local_error_buffer[SA_ARRAY_SIZE];

	DBVsa_strcpy(local_error_buffer, error_buffer);

        if (DBVcheck_dbv_close_reason(lAssocId, psCloseReq, local_error_buffer, ppsSIError))
                return SI_NOTOK;
 
        if (DBVcheck_dbv_close_resrepfmt(lAssocId, psCloseReq, local_error_buffer, ppsSIError))
                return SI_NOTOK;
 
        if (DBVcheck_dbv_close_resrep(lAssocId, psCloseReq, local_error_buffer, ppsSIError))
                return SI_NOTOK;
 
        if (DBVcheck_dbv_close_otherinfo(lAssocId, psCloseReq, local_error_buffer, ppsSIError))
                return SI_NOTOK;

	return SI_OK;
}



int DBVcheck_dbv_close_reason(DbvAssocId lAssocId, DbvCloseReq *psCloseReq, char *error_buffer, DbvError **ppsSIError)
{
	char local_error_buffer[SA_ARRAY_SIZE];

	DBVsa_strcpy(local_error_buffer, error_buffer);
	DBVsa_strcat(local_error_buffer, "eCloseReason");

       if (DBVcheck_dbv_enum(psCloseReq->eCloseReason, CR_Finished, CR_Unspecified) == SI_NOTOK) 
                return(DBVdbv_invalid_enum_error(lAssocId, local_error_buffer, ppsSIError));

	return SI_OK;
}



int DBVcheck_dbv_close_resrepfmt(DbvAssocId lAssocId, DbvCloseReq *psCloseReq, char *error_buffer, DbvError **ppsSIError)
{
	char local_error_buffer[SA_ARRAY_SIZE];

	DBVsa_strcpy(local_error_buffer, error_buffer);

        /* Can be NULL as optional */
        if (psCloseReq->pcResourceReportFormat == NULL)
		return SI_OK;

	DBVsa_strcat(local_error_buffer, "pcResourceReportFormat");

        if (DBVcheck_dbv_resrepoid(psCloseReq->pcResourceReportFormat) == SI_NOTOK)
                return(DBVdbv_oid_error(lAssocId, local_error_buffer, ppsSIError));
 
        return SI_OK;

}



int DBVcheck_dbv_close_resrep(DbvAssocId lAssocId, DbvCloseReq *psCloseReq, char *error_buffer, DbvError **ppsSIError)
{
	char local_error_buffer[SA_ARRAY_SIZE];

	DBVsa_strcpy(local_error_buffer, error_buffer);

        /* Can be NULL because optional */
        if (psCloseReq->psResourceReport == NULL)
                return SI_OK;

	DBVsa_strcat(local_error_buffer, "psResourceReport->");
        return(DBVcheck_dbv_external(lAssocId, psCloseReq->psResourceReport, local_error_buffer, ppsSIError));
 
        return SI_OK;
}



int DBVcheck_dbv_close_otherinfo(DbvAssocId lAssocId, DbvCloseReq *psCloseReq, char *error_buffer, DbvError **ppsSIError)
{
	char local_error_buffer[SA_ARRAY_SIZE];

	DBVsa_strcpy(local_error_buffer, error_buffer);

        /* Can be null because optional */
        if (psCloseReq->ppsOtherInformation == NULL)
                return SI_OK;
 
	DBVsa_strcat(local_error_buffer, "ppsOtherInformation");
        return(DBVcheck_dbv_otherinfo(lAssocId, psCloseReq->ppsOtherInformation, local_error_buffer, ppsSIError));

}




int DBVsetup_snacc_closerequest(PDU *PDU_ptr, DbvCloseReq *psCloseReq, jmp_buf env)
{
	DBVset_snacc_closereq_refid(PDU_ptr, psCloseReq->psReferenceId, env);
	DBVset_snacc_closereq_closereason(PDU_ptr, psCloseReq->eCloseReason, env);
	DBVset_snacc_closereq_diaginfo(PDU_ptr, psCloseReq->psDiagnosticInformation, env);
	DBVset_snacc_closereq_resrepform(PDU_ptr, psCloseReq->pcResourceReportFormat, env);
	DBVset_snacc_closereq_resreport(PDU_ptr, psCloseReq->psResourceReport, env);
	DBVset_snacc_closereq_otherinfo(PDU_ptr, psCloseReq->ppsOtherInformation, env);
}



int DBVset_snacc_closereq_refid(PDU *PDU_ptr, DbvOctetString *psReferenceId, jmp_buf env)
{

	/* Check if a referenceId is given(it is an optional field) */
	if  (psReferenceId != NULL) {
		DBVdbv_to_snacc_octet_string(&(PDU_ptr->a.close->referenceId), psReferenceId, env);
	}
}



int DBVset_snacc_closereq_closereason(PDU *PDU_ptr, DbvCloseReason eCloseReason, jmp_buf env)
{
	switch (eCloseReason) {
	case CR_Finished:
		PDU_ptr->a.close->closeReason = FINISHED;
		break;
	case CR_ShutDown:
		PDU_ptr->a.close->closeReason = SHUTDOWN;
		break;
	case CR_SystemProblem:
		PDU_ptr->a.close->closeReason = SYSTEMPROBLEM;
		break;
	case CR_CostLimit:
		PDU_ptr->a.close->closeReason = COSTLIMIT;
		break;
	case CR_Resources:
		PDU_ptr->a.close->closeReason = RESOURCES;
		break;
	case CR_SecurityViolation:
		PDU_ptr->a.close->closeReason = SECURITYVIOLATION;
		break;
	case CR_ProtocolError:
		PDU_ptr->a.close->closeReason = PROTOCOLERROR;
		break;
	case CR_LackOfActivity:
		PDU_ptr->a.close->closeReason = LACKOFACTIVITY;
		break;
	case CR_PeerAbort:
		PDU_ptr->a.close->closeReason = PEERABORT;
		break;
	case CR_Unspecified:
		PDU_ptr->a.close->closeReason = UNSPECIFIED;
		break;
	}
}



int DBVset_snacc_closereq_diaginfo(PDU *PDU_ptr, DbvInternationalString *psDiagnosticInformation, jmp_buf env) 
{
	if (psDiagnosticInformation != NULL)
		DBVdbv_to_snacc_inter_string(&PDU_ptr->a.close->diagnosticInformation, psDiagnosticInformation, env);
}



int DBVset_snacc_closereq_resrepform(PDU *PDU_ptr, DbvObjectIdentifier  pcResourceReportFormat, jmp_buf env)
{
	if (pcResourceReportFormat != NULL)
		ASN_MakeOid((char *) pcResourceReportFormat, &PDU_ptr->a.close->resourceReportFormat, env);
}



int DBVset_snacc_closereq_resreport(PDU *PDU_ptr, DbvASN1External *psResourceReport, jmp_buf env)
{
	if (psResourceReport != NULL) {

		/* ResourceReport is a typedefed NEWEXTERNAL */
		PDU_ptr->a.close->resourceReport = (ResourceReport *) Asn1Alloc (sizeof (ResourceReport));
		CheckAsn1Alloc(PDU_ptr->a.close->resourceReport, env);

		DBVdbv_to_snacc_external(PDU_ptr->a.close->resourceReport, psResourceReport, env);
	}
}



int DBVset_snacc_closereq_otherinfo(PDU *PDU_ptr, DbvOtherInformation **ppsOtherInformation, jmp_buf env)
{
	/* Check if ppsOtherInformation exists (it is an optional field) */
	if (ppsOtherInformation != NULL)  {
		PDU_ptr->a.close->otherInfo = AsnListNew (sizeof(void *));
		CheckAsn1Alloc(PDU_ptr->a.close->otherInfo, env);

		DBVdbv_to_snacc_otherinfo(PDU_ptr->a.close->otherInfo, ppsOtherInformation, env);
	}
}





/*****************************************************************
******************************************************************
******************************************************************
PROCESS Close PDU RECEIVED From Other End Of Connection
*/



int DBVset_dbv_closereq_refid(DbvCloseReq *closereq, PDU *PDU_ptr, jmp_buf env)
{
	/* If reference id given (it is an optional field) */
	if (PDU_ptr->a.close->referenceId.octetLen != 0){
		closereq->psReferenceId = (DbvOctetString *) DbvAlloc (sizeof (DbvOctetString));
		CheckDbvAlloc(closereq->psReferenceId, env);

		DBVsnacc_to_dbv_octet_string(closereq->psReferenceId, &(PDU_ptr->a.close->referenceId), env); 
	}
}



int DBVset_dbv_closereq_closereason(DbvCloseReq *closereq, PDU *PDU_ptr, jmp_buf env)
{
	switch (PDU_ptr->a.close->closeReason) {
	case FINISHED:
		closereq->eCloseReason = CR_Finished;
		break;
	case SHUTDOWN:
		closereq->eCloseReason = CR_ShutDown;
		break;
	case SYSTEMPROBLEM:
		closereq->eCloseReason = CR_SystemProblem;
		break;
	case COSTLIMIT:
		closereq->eCloseReason = CR_CostLimit;
		break;
	case RESOURCES:
		closereq->eCloseReason = CR_Resources;
		break;
	case SECURITYVIOLATION:
		closereq->eCloseReason = CR_SecurityViolation;
		break;
	case PROTOCOLERROR:
		closereq->eCloseReason = CR_ProtocolError;
		break;
	case LACKOFACTIVITY:
		closereq->eCloseReason = CR_LackOfActivity;
		break;
	case PEERABORT:
		closereq->eCloseReason = CR_PeerAbort;
		break;
	case UNSPECIFIED:
		closereq->eCloseReason = CR_Unspecified;
		break;
	default:
		closereq->eCloseReason = CR_Invalid;
		break;
	}
}



int DBVset_dbv_closereq_diaginfo(DbvCloseReq *closereq, PDU *PDU_ptr, jmp_buf env)
{
	if (PDU_ptr->a.close->diagnosticInformation.octetLen != 0) {
		closereq->psDiagnosticInformation = (DbvInternationalString *) DbvAlloc (sizeof (DbvInternationalString));		
		CheckDbvAlloc(closereq->psDiagnosticInformation, env);

		DBVsnacc_to_dbv_inter_string(closereq->psDiagnosticInformation, &PDU_ptr->a.close->diagnosticInformation, env);
	}
}



int DBVset_dbv_closereq_resrepform(DbvCloseReq *closereq, PDU *PDU_ptr, jmp_buf env)
{
        char *oid = NULL;
 
        if (PDU_ptr->a.close->resourceReportFormat.octetLen != 0) {
                oid = (char *) ASN_EncodedToString(&(PDU_ptr->a.close->resourceReportFormat));

		/* DbvObjectIdentifier is a typdefed char* */
                closereq->pcResourceReportFormat = (DbvObjectIdentifier) DbvAlloc (strlen(oid) + 1);
		CheckDbvAlloc(closereq->pcResourceReportFormat, env);

                memcpy(closereq->pcResourceReportFormat, oid, strlen(oid));
        }
}



int DBVset_dbv_closereq_resreport(DbvCloseReq *closereq, PDU *PDU_ptr, jmp_buf env)
{
	if (PDU_ptr->a.close->resourceReport != NULL) {
		closereq->psResourceReport = (DbvASN1External *) DbvAlloc (sizeof (DbvASN1External));
		CheckDbvAlloc(closereq->psResourceReport, env);

		DBVsnacc_to_dbv_external(closereq->psResourceReport, PDU_ptr->a.close->resourceReport, env);
	}
}



int DBVset_dbv_closereq_otherinfo(DbvCloseReq *closereq, PDU *PDU_ptr, jmp_buf env)
{
        int num_elements = 0;

	/* If the otherinformation field is given (it is an optional field) */
	if (PDU_ptr->a.close->otherInfo != NULL) {
		num_elements = AsnListCount(PDU_ptr->a.close->otherInfo);
		closereq->ppsOtherInformation = (DbvOtherInformation **) DbvAlloc (sizeof (DbvOtherInformation *) * (num_elements + 1));
		CheckDbvAlloc(closereq->ppsOtherInformation, env);

		DBVsnacc_to_dbv_otherinfo(closereq->ppsOtherInformation, PDU_ptr->a.close->otherInfo, env);
        }
}


