/*  Module Name: @(#)state_machine.c	7.1
    Version: 11/15/96
    Author: Chris Attewell
 
    Crossnet Systems Ltd.
    Unit 3, The Paddock,
    Hambridge Road,
    Newbury,
    Berkshire, RG14 5TQ
 
    Date of Issue:
 
    QA Acceptance:
    Date of Acceptance:
 
    Description: State machine code for DBV-OSI II project.
		 This file contains the functions to change the state of the 
		 association. Its shared by both origin and target.
*/

#include "state_machine.h"

#define DBV_PROTOCOL_FAILED	0

extern char DBVOriginStateErrorString[][DBV_SM_STRLEN];
extern char DBVOriginEventErrorString[][DBV_SM_STRLEN];
extern int DBVOriginStateTable[DBV_O_EVENTS][DBV_O_STATES];

extern char DBVTargetStateErrorString[][DBV_SM_STRLEN];
extern char DBVTargetEventErrorString[][DBV_SM_STRLEN];
extern int DBVTargetStateTable[DBV_T_EVENTS][DBV_T_STATES];


/**********************************
/* global variable */
/* state machine on (DBV_TRUE) or off (DBV_FALSE) */
int DBVStateMachine = DBV_TRUE;
/**********************************/


#define DBV_OACCESS_CHALLENGE_DONE 98
#define DBV_TACCESS_CHALLENGE_DONE 98


#define DBV_ORESCNTRL_CHALLENGE_DONE 99
#define DBV_TRESCNTRL_CHALLENGE_DONE 99



/* This function attempts to change the current state of the specified 
   association using the event supplied. If the state change is successful
   then the associations current state is changed. If the event is invalid
   for the current state, then the error message is completed. */

int DBVChangeState( DbvAssocId lAssocId, long lEvent, DbvError **ppsError )
{
	DbvProtocolStack	eProtocolStack;
	DBVAssociateDetails	*psAssociateDetails;
	long			lNextState;
	static long		lLastNonControlState = 0;
	int			return_value = SI_OK;


	/* If state machine off (for testing purposes) */
	if (DBVStateMachine == DBV_FALSE)
		return SI_OK;


	if(SIWhichStack(lAssocId, &eProtocolStack, ppsError) == SI_OK)
	{
		psAssociateDetails = SL_Current(DBVAssociateList);

/**************************************************************************/
		/* Origin */
		if(psAssociateDetails->eOriginOrTarget == O)
		{
			/* If just received a ResourceControlRequest PDU */
			if (lEvent == DBV_OE_RESCNTREQ_PDU)

				/* If a ResourceControlResponse PDU is not
				   required (fResponseRequired field in
				   DbvResourceControlReq set to DBV_FALSE) */
				if (ResourceControlResponseRequired == DBV_FALSE) {
					return SI_OK;
				}


			/* Get the next state to move to based on the
			   event which has just occurred */
			lNextState = DBVOriginStateTable[lEvent]
				[psAssociateDetails->lCurrentState];


			if(lNextState == DBV_PROTOCOL_FAILED ||
				lNextState == DBV_OS_ERROR)
			{
				/* Either an error from the service-user
					or the network peer has occurred */
				return_value=DBVOriginSMError(lAssocId,
					lEvent, lNextState, psAssociateDetails, 
					ppsError);
			}

			else if (lNextState == DBV_OACCESS_CHALLENGE_DONE) /* 98 (defined above and in origin-sm.c state machine table) */
			{
				/* Change back to state prior to access 
				   control request (e.g. init sent) */
				psAssociateDetails->lCurrentState = lLastNonControlState;
			}

			/* we will only ever do the following else if a 
			   ResourceControlResponse is required */
			else if (lNextState == DBV_ORESCNTRL_CHALLENGE_DONE) /* 99 (defined above and in origin-sm.c state machine table) */
			{
				/* Change back to state prior to resource 
				   control request (e.g. init sent) */
				psAssociateDetails->lCurrentState = lLastNonControlState;
			}
			else
			{
				/* Record the state in order to return to it 
				   after handling
				   access/resource control/trigger sequences */
				lLastNonControlState = psAssociateDetails->lCurrentState;

				/* Change to the new state */
				psAssociateDetails->lCurrentState =
					lNextState;

				

			}
			sprintf(DBVTraceBuffer, "ORIGIN EVENT is: %s\n", 
				DBVOriginEventErrorString[lEvent]); 
			DBVTrace(lAssocId, Trace_Low, DBVTraceBuffer);

			sprintf(DBVTraceBuffer, 
			  "Current ORIGIN STATE is: %s\n", DBVOriginStateErrorString[psAssociateDetails->lCurrentState]);
			DBVTrace(lAssocId, Trace_Low, DBVTraceBuffer);
		}

/**************************************************************************/
		/* Target */
		else
		{
			/* If want to sent a ResourceControlRequest PDU */
			if (lEvent == DBV_TE_RESCNTREQ)

				/* If a ResourceControlResponse PDU is not
				   required (fResponseRequired field in
				   DbvResourceControlReq set to DBV_FALSE) */
				if (ResourceControlResponseRequired == DBV_FALSE) {

					/* If we have previously sent a
					   ResourceControlRequest PDU with 
					   TRUE its an error ! */
					if (psAssociateDetails->lCurrentState == DBV_TS_RESCNT_SENT) {
						lNextState = DBVTargetStateTable[lEvent] [psAssociateDetails->lCurrentState];
						return(DBVTargetSMError(lAssocId, lEvent, lNextState, psAssociateDetails, ppsError));
					} else
						return SI_OK;
				}


			/* The Supplied association is a target */
			lNextState = DBVTargetStateTable[lEvent]
				[psAssociateDetails->lCurrentState];


			if(lNextState == DBV_PROTOCOL_FAILED ||
				lNextState == DBV_TS_ERROR)
			{
				/* Either an error from the service-user
					or the network peer has occurred */
				return_value=DBVTargetSMError(lAssocId,
					lEvent, lNextState, psAssociateDetails, 
					ppsError);
			}


			else if (lNextState == DBV_TACCESS_CHALLENGE_DONE) /* 98 (defined above and in target-sm.c state machine table) */
			{
				/* Change back to state prior to access 
				   control request (e.g. init sent) */
				psAssociateDetails->lCurrentState = lLastNonControlState;
			}

			/* we will only ever do the following else if a 
			   ResourceControlResponse is required */
			else if (lNextState == DBV_TRESCNTRL_CHALLENGE_DONE) /* 99 (defined above and in target-sm.c state machine table) */
			{
				/* Change back to state prior to resource 
				   control request 
				   (e.g. init sent) */
				psAssociateDetails->lCurrentState = lLastNonControlState;
			}

			else
			{

				/* Record the state in order to return to it 
				   after handling access/resource control/
				   trigger sequences */
				lLastNonControlState = psAssociateDetails->lCurrentState;
				
				psAssociateDetails->lCurrentState =
					lNextState;
			}
			sprintf(DBVTraceBuffer, "TARGET EVENT is: %s\n", 
				DBVTargetEventErrorString[lEvent]); 
			DBVTrace(lAssocId, Trace_Low, DBVTraceBuffer);

			sprintf(DBVTraceBuffer, 
			  "Current TARGET STATE is: %s\n", DBVTargetStateErrorString[psAssociateDetails->lCurrentState]);
			DBVTrace(lAssocId, Trace_Low, DBVTraceBuffer);
		}

/**************************************************************************/

	
	}
	else
		/* Error completed by which stack */
		return_value = SI_NOTOK;

	return return_value;
}



/* Set the initial state for the origin */
void DBVSetState(DbvAssocId lAssocId, long lNewState)
{
	DbvError		*psError=NULL;
	DbvProtocolStack         eProtocolStack;
	DBVAssociateDetails     *psAssociateDetails;

	if(SIWhichStack(lAssocId, &eProtocolStack, &psError) == SI_OK)
	{
		psAssociateDetails = SL_Current(DBVAssociateList);
		psAssociateDetails->lCurrentState = lNewState;
	}
}



/* Get the current state of the target */
long DBVGetTargetState(DbvAssocId lAssocId)
{
        DbvError                *psError=NULL;
        DbvProtocolStack         eProtocolStack;
        DBVAssociateDetails     *psAssociateDetails;
 
        if(SIWhichStack(lAssocId, &eProtocolStack, &psError) == SI_OK)
        {
                psAssociateDetails = SL_Current(DBVAssociateList);
		return psAssociateDetails->lCurrentState;
        }
	else
		return DBV_TS_ERROR;
}



/* Generate a state machine error for the origin */
int DBVOriginSMError(DbvAssocId lAssocId, long lEvent, long lNextState, 
	DBVAssociateDetails *psAssociateDetails, DbvError **ppsError)
{
	int return_value = SI_NOTOK;

	if(lNextState == DBV_OS_ERROR)
	{
		/* Incoming event (from the network) has caused us to go into
			the error state */
		sprintf(DBVErrorBuffer, DBV_SM_ERROR_S, 
			DBVOriginEventErrorString[lEvent], 
			DBVOriginStateErrorString[psAssociateDetails->
				lCurrentState]);
		DBVfill_error_structure(lAssocId, SI_Interface_Error, 0,
			DBV_SM_ERROR, DBVErrorBuffer, 0, Association_Open, 
			ppsError);
		
		/* Move to the error state */
		psAssociateDetails->lCurrentState =
			lNextState;
	}
	else
	{
		/* The service-user has been naughty */
		sprintf(DBVErrorBuffer, DBV_SM_VIOLATION_S, 
			DBVOriginEventErrorString[lEvent], 
			DBVOriginStateErrorString[psAssociateDetails->
				lCurrentState]);
		DBVfill_error_structure(lAssocId, SI_Interface_Error, 0,
			DBV_SM_VIOLATION, DBVErrorBuffer, 0, Association_Open, 
			ppsError);
	}

	return return_value;
}



/* Generates a state machine error for the target */
int DBVTargetSMError(DbvAssocId lAssocId, long lEvent, long lNextState,
	DBVAssociateDetails *psAssociateDetails, DbvError **ppsError)
{
	int return_value = SI_NOTOK;

	if(lNextState == DBV_TS_ERROR)
	{
		/* Incoming event (from the network) has caused us to go into
			the error state */
		sprintf(DBVErrorBuffer, DBV_SM_ERROR_S, 
			DBVTargetEventErrorString[lEvent], 
			DBVTargetStateErrorString[psAssociateDetails->
				lCurrentState]);
		DBVfill_error_structure(lAssocId, SI_Interface_Error, 0,
			DBV_SM_ERROR, DBVErrorBuffer, 0, Association_Open, 
			ppsError);
		
		/* Move to the error state */
		psAssociateDetails->lCurrentState =
			lNextState;
	}
	else
	{
		/* The service-user has been naughty */
		sprintf(DBVErrorBuffer, DBV_SM_VIOLATION_S, 
			DBVTargetEventErrorString[lEvent], 
			DBVTargetStateErrorString[psAssociateDetails->
				lCurrentState]);
		DBVfill_error_structure(lAssocId, SI_Interface_Error, 0,
			DBV_SM_VIOLATION, DBVErrorBuffer, 0, Association_Open, 
			ppsError);
	}

	return return_value;
}



