/*  Module Name: @(#)tbl.c	1.3
    Version: 2/26/97
    Author:  M. Bush
    
    Crossnet Systems Ltd.
    Unit 3, The Paddock,
    Hambridge Road,
    Newbury,
    Berkshire, RG14 5TQ

    Date of Issue: 

    QA Acceptance: 
    Date of Acceptance: 

    Description:  
		  TBL object source code.
	Changes to allow appending to tbl files and punctuation (except "," or 
	";") in the unquoted text field by Chris Attewell
*/

#include <ctype.h>
#include <stdio.h>
#include <math.h>
#include "memcheck.h"
#include "tbl.h"
#include "sl_list.h"
#include "str_hand.h"


/*
 * Returns a pointer to a tabular file reading and parsing (TBL) object. The 'Filename' 
 * argument is a pointer to a character string containing the name of the TBL file
 * to be read.
 */

TBL	TBL_Create(char*Filename, TBL_Mode Mode)
{
	TBL	    Table;
	char	Message[400];

	Table = SMALLOC_(struct tbl_Table , "TBL");
	Table->FieldParseTree = SL_Create( (SL_Free)tbl_ThrowToken);
	if (Table->FieldParseTree == NULL) {
		SError_(Fatal_, "SL Create failed", "TBL");
	} else {
		switch ( (int)Mode) {
		case (int)
TBL_Write :
				Table->Stream = 
				    fopen(Filename, "w");
			if (Table->Stream == NULL) {
				sprintf(Message, "Unable to open file %s\n", Filename);
				return NULL;
			}
			break;
		case (int)
TBL_Read :
				Table->Stream = 
				    fopen(Filename, "r");
			if (Table->Stream == NULL) {
				sprintf(Message, "Unable to open file %s\n", Filename);
				return NULL;
			}
			break;
/* Append mode added by cta */
		case (int)
TBL_Append:
                                Table->Stream =
                                    fopen(Filename, "a");
                        if (Table->Stream == NULL) {
                                sprintf(Message, "Unable to open file %s\n", Filename);
                                return NULL;
                        }
			break;

		default :
			SError_(Fatal_, "Invalid mode", "TBL");
		};
		if (Table->Stream == NULL) {
			return(NULL);
		} else {
			return(Table);
		}
	}
	return NULL;
}

/*
 * De-allocates memory used by the TBL object.
 */

void    TBL_Destruct(TBL Table)
{
	fclose(Table->Stream);
	SL_Destruct(Table->FieldParseTree);
	FREE_(Table);
}


/*
 * Reads the next record from a TBL file. A record is defined as a string of
 * characters (with those encased between comment markers ignored) which is 
 * terminated by a semi-colon. Returns False if unable to read a record or has come to 
 * the end of the file, otherwise returns True.
 */


Bool_	TBL_GetRecord(TBL Table)
{
	tbl_TokenPtr    AToken;
	Bool_	    SyntaxError = False;

	tbl_ClearBuffer(Table);
	if (tbl_ParseFieldIntoTree(Table) == True) {
		AToken = tbl_Token(Table); /* Get field or record sep */
		if ( !( (AToken->Type == tbl_FieldSep) || 
		    (AToken->Type == tbl_RecTerm) ) ) {
			/* SyntaxError */
			return(False);
		} else {
			while ( (AToken->Type == tbl_FieldSep) && (SyntaxError == False) ) {
				if (tbl_ParseFieldIntoTree(Table) == True) {
					tbl_ThrowToken(AToken);
					AToken = tbl_Token(Table);  /* get field or record sep */
					if ( !( (AToken->Type == tbl_FieldSep) || 
					    (AToken->Type == tbl_RecTerm) ) ) {
						SyntaxError = True;
					}
				} else {
					SyntaxError = True;
				}
			};
			if (AToken->Type == tbl_RecTerm && !SyntaxError) {
				tbl_ThrowToken(AToken);
				return(True);
			} else {
				tbl_ThrowToken(AToken);
				return(False);
			}
		}
	} else {
		return (False);
	}
}








/*
 * Writes the next record to a TBL file. A record is defined as a string of characters 
 * (with those encased between comment markers ignored) which is terminated by a 
 * semi-colon. Returns FALSE if unable to write a record, otherwise returns TRUE.
 */

Bool_	TBL_PutRecord(TBL Table)
{
	tbl_TokenPtr    AToken;
	tbl_TokenType   PreviousType;

	AToken = (tbl_TokenPtr)SL_Remove(Table->FieldParseTree);
	if ( AToken != NULL ) {
		tbl_WriteField(Table->Stream, AToken);
		PreviousType = AToken->Type;
		tbl_ThrowToken(AToken);
		AToken = (tbl_TokenPtr)SL_Remove(Table->FieldParseTree);
		while ( AToken != NULL ) {
			if (PreviousType != tbl_Comment) {
				fputc(FIELD_SEP, Table->Stream);
			};
			tbl_WriteField(Table->Stream, AToken);
			PreviousType = AToken->Type;
			tbl_ThrowToken(AToken);
			AToken = (tbl_TokenPtr)SL_Remove(Table->FieldParseTree);
		};
		fputc(RECORD_TERM, Table->Stream);
		fputc('\n', Table->Stream);
		return( True );
	} else {
		return (False);
	};
}


/*
 * Returns the next field in a record in the TBL_Return structure.  Fields are
 * separated by commas within records. Returns False if unable to interpret a field or 
 * has come to the end of a record, otherwise returns True.
 */

Bool_   TBL_GetField(TBL Table, TBL_Return Return)
{
	tbl_TokenPtr    AToken;

	AToken = (tbl_TokenPtr)SL_Remove(Table->FieldParseTree);
	if (AToken == NULL) {
		return(False);
	} else {
		switch (AToken->Type) {
		case tbl_Int:
			Return->Tag = TBL_Int;
			Return->Value.Int = AToken->Value.AnInt;
			break;
		case tbl_Double:
			Return->Tag = TBL_Double;
			Return->Value.Double = AToken->Value.ADouble;
			break;
		case tbl_QText:
			Return->Tag = TBL_QText;
			Return->Value.QText = STR_Save(AToken->Value.QText);
			Return->Value.Text = Return->Value.QText;
			break;
		case tbl_UText:
			Return->Tag = TBL_UText;
			Return->Value.UText = STR_Save(AToken->Value.UText);
			Return->Value.Text = Return->Value.UText;
			break;
		case tbl_NullField:
			Return->Tag = TBL_NullField;
			break;
		default:
			return(False);
		}
		tbl_ThrowToken(AToken);
		return(True);
	}
}


/*
 * Writes the next field in a record in the TBL_Return structure.  Fields are
 * separated by commas within records. Returns False if unable to write a field, 
 * otherwise returns True.
 */

Bool_   TBL_PutField(TBL Table, TBL_Return Return)
{
	tbl_TokenPtr    AToken = SMALLOC_(struct tbl_Token , "TBL");

	switch (Return->Tag) {
	case TBL_Int:
		AToken->Type = tbl_Int;
		AToken->Value.AnInt = Return->Value.Int;
		break;
	case TBL_Double:
		AToken->Type = tbl_Double;
		AToken->Value.ADouble = Return->Value.Double;
		break;
	case TBL_QText:
		AToken->Type = tbl_QText;
		AToken->Value.QText = STR_Save(Return->Value.QText);
		break;
	case TBL_UText:
		AToken->Type = tbl_UText;
		AToken->Value.UText = STR_Save(Return->Value.UText);
		break;
	case TBL_Comment:
		AToken->Type = tbl_Comment;
		AToken->Value.Comment = STR_Save(Return->Value.Comment);
		break;
	case TBL_NullField:
		AToken->Type = tbl_NullField;
		Return->Tag = TBL_NullField;
		break;
	default:
		return(False);
	};
	SL_Add(Table->FieldParseTree, AToken);
	return(True);
}


/*
 * Returns a pointer to a token structure containing the next token from the TBL file 
 * currently being read. Memory is dynamically allocated to the tbl_TokenPtr
 * within this function. Called by TBL_GetRecord.
 */

tbl_TokenPtr tbl_Token(TBL Table)
{
	int	AChar;
	Bool_	    WhiteSpace;
	Bool_	    Comment;
	char	TestChar;

	tbl_TokenPtr    AToken = SMALLOC_(struct tbl_Token , "TBL");
	do {			    /* Strip comments and whitespace */
		WhiteSpace = False;
		Comment	 = False;
		AChar = fgetc(Table->Stream);
		if (isspace(AChar)) {
			WhiteSpace = True;
		} else if (AChar == COMMENT_START) {
			tbl_GetComment(Table->Stream);
			Comment = True;
		};
	} while (WhiteSpace == True || Comment == True);
	if (AChar == '"') {
		AToken->Type = tbl_QText;
		AToken->Value.QText = tbl_GetQText(Table->Stream);
	} else if (isalpha(AChar)) {
		if (ungetc(AChar, Table->Stream) != AChar) {
			SError_(Fatal_, "Cannot ungetc", "TBL");
		}
		AToken->Type = tbl_UText;
		AToken->Value.UText = tbl_GetUText(Table->Stream);
	} else if ( (AChar == '+') || (AChar == '-') || (isdigit(AChar)) ) {
		TestChar = ungetc(AChar, Table->Stream);
		if ( TestChar != AChar) {
			SError_(Fatal_, "Cannot ungetc", "TBL");
		}
		tbl_GetNumber(Table->Stream, AToken);
	} else if (AChar == FIELD_SEP) {
		AToken->Type = tbl_FieldSep;
	} else if (AChar == RECORD_TERM) {
		AToken->Type = tbl_RecTerm;
	} else if (AChar == FILE_TERM) {
		AToken->Type = tbl_FileTerm;
	} else if (AChar == NULL_FIELD) {
		AToken->Type = tbl_NullField;
	} else {
		AToken->Type = tbl_TokenError;
	};
	return(AToken);
}


/*
 * Moves the file reader to the end of a comment. Called by tbl_token.
 */

Global_	void	tbl_GetComment(FILE*Stream)
{
	char	AChar1 = fgetc(Stream);
	char	AChar2 = fgetc(Stream);

	while (!(AChar1 == '*' && AChar2 == '/')) {
		AChar1 = AChar2;
		AChar2 = fgetc(Stream);
	};
}


/*
 * Reads an integer number from the file and places and returns in 'AToken'.
 */

Global_	void	    tbl_GetNumber(FILE*Stream, tbl_TokenPtr AToken)
{
	long int	Integral;
	char	DecimalPoint;
	double	DoublePart;
	int	TestItems;

	TestItems = fscanf(Stream, "%ld%c", &Integral, &DecimalPoint);
	if ( TestItems == EOF) {
		AToken->Type = tbl_TokenError;
	} else {
		if (DecimalPoint == '.') {
			if (ungetc(DecimalPoint, Stream) != DecimalPoint) {
				SError_(Fatal_, "Cannot ungetc", "TBL");
			}
			fscanf(Stream, "%lf", &DoublePart);
			if (Integral >= 0) {
				AToken->Value.ADouble = DoublePart + (double)Integral;
			} else {
				AToken->Value.ADouble = (double)Integral - DoublePart;
			}
			AToken->Type = tbl_Double;
		} else {
			AToken->Value.AnInt = Integral;
			if (ungetc(DecimalPoint, Stream) != DecimalPoint) {
				SError_(Fatal_, "Cannot ungetc", "TBL");
			}
			AToken->Type = tbl_Int;
		}
	}
}


/*
 * Reads an unquoted text field from the file and returns a pointer to the read text 
 * string. Called by tbl_token.
 */

char*	tbl_GetUText(FILE*Stream)
{
	char	AChar;
	char	StoreBuffer[STORE_BUFFER_SIZE];
	int	StoreCounter = 0;
	char * 	AString;

	for (AChar = (char)fgetc(Stream); 
	  /*  isalnum(AChar); removed by cta for DBV-OSI II */
	      AChar != ',' && AChar != ';';
	    AChar = (char)fgetc(Stream)) {
		StoreBuffer[StoreCounter] = AChar;
		StoreCounter++;
	};
	if (ungetc(AChar, Stream) != AChar) {
		SError_(Fatal_, "Cannot ungetc", "TBL");
	}
	StoreBuffer[StoreCounter] = '\0';
	AString = STR_Save(StoreBuffer);
	return(AString);
}


/*
 * Reads a quoted text field from the file, strips out the quotation marks, and
 * returns a pointer to the read text string. Called by tbl_token.
 */

char*	tbl_GetQText(FILE*Stream)
{
	char	AChar1, AChar2;
	char	StoreBuffer[STORE_BUFFER_SIZE];
	int	StoreCounter = 0;
	char * 	AString;

	AChar1 = fgetc(Stream);
	AChar2 = fgetc(Stream);
	while ( ((AChar1 == '"' && AChar2 == '"') || (AChar1 != '"')) && 
	    ((AChar1 != (char)FILE_TERM) && (AChar2 != (char)FILE_TERM)) ) {
		if (AChar1 == '"') {
			StoreBuffer[StoreCounter] = '"';
			StoreCounter++;
			AChar1 = fgetc(Stream);
			AChar2 = fgetc(Stream);
		} else {
			StoreBuffer[StoreCounter] = AChar1;
			StoreCounter++;
			AChar1 = AChar2;
			AChar2 = fgetc(Stream);
		};
	};
	StoreBuffer[StoreCounter] = '\0';
	if (ungetc(AChar2, Stream) != AChar2) {
		SError_(Fatal_, "Cannot ungetc", "TBL");
	}
	AString = STR_Save(StoreBuffer);
	return(AString);
}


/*
 * Removes all tokens from the 'SL' list added to on calling tbl_ParseFieldIntoTree.
 */

Global_	void		tbl_ClearBuffer(TBL Table)
{
	tbl_TokenPtr AToken;

	for (AToken = (tbl_TokenPtr)SL_Remove(Table->FieldParseTree); 
	    AToken != NULL; 
	    AToken = (tbl_TokenPtr)SL_Remove(Table->FieldParseTree)) {
		tbl_ThrowToken(AToken);
	}
}


/*
 * Retrieves the next token and places into an 'SL' List if it is a field type, 
 * otherwise it de-allocates memory allocated to the token. Called by TBL_GetRecord
 * and tbl_token.
 */


Global_	Bool_	tbl_ParseFieldIntoTree(TBL Table)
{
	tbl_TokenPtr    AToken;

	AToken = tbl_Token(Table);
	switch (AToken->Type) {
	case tbl_NullField:
	case tbl_QText:
	case tbl_UText:
	case tbl_Int:
	case tbl_Double:
		SL_Add(Table->FieldParseTree, AToken);
		return(True);
	default:
		tbl_ThrowToken(AToken);
		return(False);
	};
}


/*
 * De-allocates memory dynamically allocated to a token.
 */

Global_	void	tbl_ThrowToken(tbl_TokenPtr AToken)
{
	switch (AToken->Type) {
	case tbl_QText:
		FREE_(AToken->Value.QText);
		break;
	case tbl_UText:
		FREE_(AToken->Value.UText);
		break;
	case tbl_Comment:
		FREE_(AToken->Value.Comment);
		break;
	default:
		break;
	};

	FREE_(AToken);
}


/*
 * Writes a field in 'AToken' to a file  with file pointer 'Stream'.
 */

Global_ void tbl_WriteField(FILE *Stream, tbl_TokenPtr AToken)
{
	switch (AToken->Type) {
	case tbl_Int:
		fprintf(Stream, "%d", AToken->Value.AnInt);
		break;
	case tbl_Double:
		fprintf(Stream, "%f", AToken->Value.ADouble);
		break;
	case tbl_UText:
		fprintf(Stream, "%s", AToken->Value.UText);
		break;
	case tbl_QText:
		fputc('"', Stream);
		fprintf(Stream, "%s", AToken->Value.QText);
		fputc('"', Stream);
		break;
	case tbl_Comment:
		fputc('/', Stream);
		fputc('*', Stream);
		fprintf(Stream, "%s", AToken->Value.Comment);
		fputc('*', Stream);
		fputc('/', Stream);
		fputc('\n', Stream);
		break;
	case tbl_NullField:
		fputc(NULL_FIELD, Stream);
		break;
	default:
		SError_(Fatal_, "Invalid field type", "TBL");
	};
}


